'''
# Flybys in `REBOUND`
- A package for implementing flybys in [hannorein/rebound](https://github.com/hannorein/rebound)
'''
from .flybys import *
from .environments import *
from .tools import *
from .analytic import *
from .imf import IMF
from . import imf
from .stars import Star,Stars
from . import stars

__version__ = 'v0.4.3'


import numpy as _numpy
import rebound as _rebound
from scipy.special import j0 as _j0,jv as _jv

from .tools import *
from . import units

############################################################
################## Analytical Estimates ####################
############################################################

def binary_energy(sim):#sun_mass, planet_mass, planet_a):
    '''
        The energy of a binary system, -(G*M*m)/(2*a).
        
        Parameters
        ----------
        sim : REBOUND Simulation
    '''
    unit_set = rebound_units(sim)
    G = (sim.G * unit_set['length']**3 / unit_set['mass'] / unit_set['time']**2)
    p = sim.particles
    return (-G * p[0].m * unit_set['mass'] * p[1].m * unit_set['mass'] / (2. * p[1].a * unit_set['length'])).decompose(list(unit_set.values()))
    
def energy_change_adiabatic_estimate(sim, star, averaged=True):
    '''
        An analytical estimate for the change in energy of a binary system due to a flyby star.
        
        From the conclusions of Roy & Haddow (2003) https://ui.adsabs.harvard.edu/abs/2003CeMDA..87..411R/abstract and Heggie (2006) https://ui.adsabs.harvard.edu/abs/2006fbp..book...20H/abstract. The orbital element angles of the flyby star are determined with respect to the plane defined by the binary orbit. In REBOUND this is the same as when the inclination of the planet is zero.
        
        Parameters
        ----------
        sim : REBOUND Simulation
        star : AIRBALL Star flyby object
    '''
    unit_set = rebound_units(sim)
    t0 = 0*unit_set['time']
    G = (sim.G * unit_set['length']**3 / unit_set['mass'] / unit_set['time']**2)

    sim = sim.copy()
    sim.rotate(_rebound.Rotation.to_new_axes(newz=sim.angular_momentum()))
    
    p = sim.particles
    m1, m2, m3 = p[0].m * unit_set['mass'], p[1].m * unit_set['mass'], star.mass # redefine the masses for convenience
    M12 = m1 + m2 # total mass of the binary system
    M123 = m1 + m2 + m3 # total mass of all the objects involved
    
    mu = G*M123
    es = vinf_and_b_to_e(mu=mu, star_b=star.b, star_vinf=star.v)
    
    a, e = p[1].a * unit_set['length'], p[1].e # redefine the orbital elements of the planet for convenience
    n = _numpy.sqrt(G*M12/a**3) # compute the mean motion of the planet
    
    w, W, i = star.omega, star.Omega, star.inc # redefine the orientation elements of the flyby star for convenience
    V = star.v
    GM123 = G*M123 
    q = (- GM123 + _numpy.sqrt( GM123**2. + star.b**2. * V**4.))/V**2. # compute the periapsis of the flyby star

    # Calculate the following convenient functions of the planet's eccentricity and Bessel functions of the first kind of order n.
    e1 = _jv(-1,e) - 2*e*_j0(e) + 2*e*_jv(2,0) - _jv(3,e)
    e2 = _jv(-1,e) - _jv(3,e)
    e4 = _jv(-1,e) - e*_j0(e) - e*_jv(2,e) + _jv(3,e)

    # Calculate a convenient function of the planet's semi-major axis and the flyby star's periapsis.
    k = _numpy.sqrt((2*M12*q**3)/(M123*a**3))
    
    # Calculate convenient functions of the flyby star's eccentricity.
    f1 = ((es + 1.0)**(0.75)) / ((2.0**(0.75)) * (es*es))
    with units.set_enabled_equivalencies(units.dimensionless_angles()):
        f2 = (3.0/(2.0*_numpy.sqrt(2.0))) * (_numpy.sqrt((es*es) - 1.0) - _numpy.arccos(1.0/es)) / ((es - 1.0)**(1.5))
    
    # Compute the prefactor and terms of the calculation done by Roy & Haddow (2003)
    prefactor = (-_numpy.sqrt(_numpy.pi)/8.0) * ((G*m1*m2*m3)/M12) * ((a*a)/(q*q*q)) * f1 * k**(2.5) * _numpy.exp((-2.0*k/3.0)*f2)
    with units.set_enabled_equivalencies(units.dimensionless_angles()):
        term1 = e1 * ( _numpy.sin(2.0*w + n*t0)*_numpy.cos(2.0*i - 1.0)- _numpy.sin(2.0*w + n*t0)*_numpy.cos(2.0*i)*_numpy.cos(2.0*W) - 3.0*_numpy.sin(n*t0 + 2.0*w)*_numpy.cos(2.0*W) - 4.0*_numpy.sin(2.0*W)*_numpy.cos(2.0*w + n*t0)*_numpy.cos(i) )
        term2 = e2 * (1.0 - e*e) * ( _numpy.sin(2.0*w + n*t0)*(1.0-_numpy.cos(2.0*i)) - _numpy.sin(2.0*w + n*t0)*_numpy.cos(2.0*i)*_numpy.cos(2.0*W) - 3.0*_numpy.sin(n*t0 +2.0*w)*_numpy.cos(2.0*W) - 4.0*_numpy.cos(n*t0 + 2.0*w)*_numpy.sin(2.0*W)*_numpy.cos(i) )
        term3 = e4 * _numpy.sqrt(1.0 - e*e) * (-2.0*_numpy.cos(2.0*i)*_numpy.cos(2.0*w + n*t0)*_numpy.sin(2.0*W) - 6.0*_numpy.cos(2.0*w + n*t0)*_numpy.sin(2.0*W) - 8.0*_numpy.cos(2.0*W)*_numpy.sin(2.0*w + n*t0)*_numpy.cos(i) )
    
    if averaged: return (prefactor).decompose(list(unit_set.values()))
    else: return (prefactor * ( term1 + term2 + term3)).decompose(list(unit_set.values()))

def relative_energy_change(sim, star, averaged=False):
    '''
        An analytical estimate for the relative change in energy of a binary system due to a flyby star.
        
        Combines energy_change_adiabatic_estimate(...) and binary_energy(...) functions.
        
        Parameters
        ----------
        sim : REBOUND Simulation with two bodies, a central star and a planet
        star : AIRBALL Star flyby object
    '''
    return energy_change_adiabatic_estimate(sim=sim, star=star, averaged=averaged)/binary_energy(sim)

def eccentricity_change_adiabatic_estimate(sim, star, averaged=False):
    '''
        An analytical estimate for the change in eccentricity of an eccentric binary system due to a flyby star.
        
        From the conclusions of Heggie & Rasio (1996) Equation (A3) from Spurzem et al. (2009) https://ui.adsabs.harvard.edu/abs/2009ApJ...697..458S/abstract. 
        The orbital element angles of the flyby star are determined with respect to the plane defined by the binary orbit. In REBOUND this is the same as when the inclination of the planet is zero.
        
        Parameters
        ----------
        sim : REBOUND Simulation with two bodies, a central star and a planet
        star : AIRBALL Star flyby object
    '''


    unit_set = rebound_units(sim)
    t0 = 0*unit_set['time']
    G = (sim.G * unit_set['length']**3 / unit_set['mass'] / unit_set['time']**2)
    
    p = sim.particles
    m1, m2, m3 = p[0].m * unit_set['mass'], p[1].m * unit_set['mass'], star.mass # redefine the masses for convenience
    M12 = m1 + m2 # total mass of the binary system
    M123 = m1 + m2 + m3 # total mass of all the objects involved
    
    mu = G*M123
    es = vinf_and_b_to_e(mu=mu, star_b=star.b, star_vinf=star.v)
    
    a, e = p[1].a * unit_set['length'], p[1].e # redefine the orbital elements of the planet for convenience
    n = _numpy.sqrt(G*M12/a**3) # compute the mean motion of the planet
    
    w, W, i = star.omega, star.Omega, star.inc # redefine the orientation elements of the flyby star for convenience
    V = star.v
    GM123 = G*M123 
    q = (- GM123 + _numpy.sqrt( GM123**2. + star.b**2. * V**4.))/V**2. # compute the periapsis of the flyby star
    
    prefactor = (-15.0/4.0) * m3 / _numpy.sqrt(M12*M123) * ((a/q)**1.5) * ((e * _numpy.sqrt(1.0 - e*e))/((1.0 + es)**1.5))
    with units.set_enabled_equivalencies(units.dimensionless_angles()):
        t1 = (_numpy.sin(i) * _numpy.sin(i) * _numpy.sin(2.0*W) * ( _numpy.arccos(-1.0/es) + _numpy.sqrt(es*es - 1.0) )).value
        t2 = ((1.0/3.0) * (1.0 + _numpy.cos(i)*_numpy.cos(i)) * _numpy.cos(2.0*w) * _numpy.sin(2.0*W)).value
        t3 = (2.0 * _numpy.cos(i) * _numpy.sin(2.0*w) * _numpy.cos(2.0*W) * ((es*es - 1.0)**1.5)/(es*es)).value

    if averaged: return (prefactor * ( ( _numpy.arccos(-1.0/es).value + _numpy.sqrt(es*es - 1.0) ) + (2.0/3.0) + (2.0 * ((es*es - 1.0)**1.5)/(es*es)) )).decompose(list(unit_set.values()))
    else: return (prefactor * (t1 + t2 + t3)).decompose(list(unit_set.values()))

def energy_change_close_encounters_sim(sim):
    '''
        An analytical estimate for the change in energy of a binary system due to a flyby star.
        
        From the conclusions of Roy & Haddow (2003) https://ui.adsabs.harvard.edu/abs/2003CeMDA..87..411R/abstract and Heggie (2006) https://ui.adsabs.harvard.edu/abs/2006fbp..book...20H/abstract. The orbital element angles of the flyby star are determined with respect to the plane defined by the binary orbit. In REBOUND this is the same as when the inclination of the planet is zero.
        
        Parameters
        ----------
        sim : three-body REBOUND sim
    '''
    s = sim.copy()
    s.move_to_hel()
    p = s.particles

    G = s.G # Newton's Gravitational constant in units of Msun, AU, Yr2Pi
    m1, m2, m3 = p[0].m, p[1].m, p[2].m # redefine the masses for convenience
    M12 = m1 + m2 # total mass of the binary system
    M23 = m2 + m3 # total mass of the second and third bodies
    M123 = m1 + m2 + m3 # total mass of all the objects involved

    V = p[2].v # velocity of the star
    es = p[2].e
    GM123 = G*M123
    b = -p[2].a * _numpy.sqrt(es*es - 1.0)
    q = (- GM123 + _numpy.sqrt( GM123**2. + b**2. * V**4.))/V**2. # compute the periapsis of the flyby star

    x,y,z = p[2].xyz
    vx,vy,vz = p[1].vxyz

    cosϕ = 1.0/_numpy.sqrt(1.0 + ((q**2.0)*(V**4.0))/(M23**2.0))
    
    prefactor = (-2.0 * m1 * m2 * m3)/(M12 * M23) * V * cosϕ
    t1 = -(x*vx + y*vy + z*vz)
    t2 = (m3 * V * cosϕ)/M23
    
    sim.move_to_hel()
    
    return prefactor * (t1 + t2)



import numpy as _numpy
from scipy.stats import uniform as _uniform
from .tools import *

try:
    # Required for Python>=3.9
    from collections.abc import MutableMapping
except:
    from collections import MutableMapping


class Star:
  '''
    This is the AIRBALL Star class.
    It encapsulates the relevant parameters for a given star.
    Only the mass is an quantity intrinsic to the object.
    The impact parameter, velocity, inclination, argument of periastron, and longitude of the ascending node quantities are defined with respect to the host star and plane passing through the star.
  '''
  def __init__(self, m, b, v, inc=None, omega=None, Omega=None, UNIT_SYSTEM=[]) -> None:
    '''
    The Mass, m, (Msun), Impact Parameter, b, (AU), Velocity, v, (km/s)
    Inclination, inc, (rad), Argument of the Periastron, omega (rad), and Longitude of the Ascending Node, Omega, (rad)
    Or define an astropy.units system, i.e. UNIT_SYSTEM = [u.pc, u.Myr, u.solMass, u.rad, u.km/u.s].
    '''
    self.units = UnitSystem(UNIT_SYSTEM)

    if inc == 'uniform' or inc == None: inc = 2.0*_numpy.pi * _uniform.rvs() - _numpy.pi
    if omega == 'uniform' or omega == None: omega = 2.0*_numpy.pi * _uniform.rvs() - _numpy.pi
    if Omega == 'uniform' or Omega == None: Omega = 2.0*_numpy.pi * _uniform.rvs() - _numpy.pi

    self.mass = m
    self.impact_parameter = b
    self.velocity = v
    self.inc = inc
    self.argument_periastron = omega
    self.longitude_ascending_node = Omega

  @property
  def UNIT_SYSTEM(self):
    return self.units.UNIT_SYSTEM

  @UNIT_SYSTEM.setter
  def UNIT_SYSTEM(self, UNIT_SYSTEM):
    self.units.UNIT_SYSTEM = UNIT_SYSTEM

  @property
  def m(self):
    return self._mass.to(self.units.units['mass'])

  @m.setter
  def m(self, value):
    self._mass = value.to(self.units.units['mass']) if isQuantity(value) else value * self.units.units['mass']

  @property
  def mass(self):
    return self.m

  @mass.setter
  def mass(self, value):
    self.m = value
  
  @property
  def b(self):
    return self._impact_parameter.to(self.units.units['length'])

  @b.setter
  def b(self, value):
    self._impact_parameter = value.to(self.units.units['length']) if isQuantity(value) else value * self.units.units['length']

  @property
  def impact_parameter(self):
    return self.b

  @impact_parameter.setter
  def impact_parameter(self, value):
    self.b = value

  @property
  def v(self):
    return self._velocity.to(self.units.units['velocity'])

  @v.setter
  def v(self, value):
    self._velocity = value.to(self.units.units['velocity']) if isQuantity(value) else value * self.units.units['velocity']

  @property
  def velocity(self):
    return self.v

  @velocity.setter
  def velocity(self, value):
    self.v = value

  @property
  def inc(self):
    return self._inclination.to(self.units.units['angle'])

  @inc.setter
  def inc(self, value):
    self._inclination = value.to(self.units.units['angle']) if isQuantity(value) else value * self.units.units['angle']

  @property
  def inclination(self):
    return self.inc

  @inc.setter
  def inclination(self, value):
    self.inc = value

  @property
  def omega(self):
    return self._argument_periastron.to(self.units.units['angle'])

  @omega.setter
  def omega(self, value):
    self._argument_periastron = value.to(self.units.units['angle']) if isQuantity(value) else value * self.units.units['angle']

  @property
  def argument_periastron(self):
    return self.omega

  @argument_periastron.setter
  def argument_periastron(self, value):
    self.omega = value

  @property
  def Omega(self):
    return self._longitude_ascending_node.to(self.units.units['angle'])

  @Omega.setter
  def Omega(self, value):
    self._longitude_ascending_node = value.to(self.units.units['angle']) if isQuantity(value) else value * self.units.units['angle']

  @property
  def longitude_ascending_node(self):
    return self.Omega

  @longitude_ascending_node.setter
  def longitude_ascending_node(self, value):
    self.Omega = value

  @property
  def params(self):
    return [self.m, self.b, self.v, self.inc, self.omega, self.Omega]
  
  @property
  def param_values(self):
    return _numpy.array([self.m.value, self.b.value, self.v.value, self.inc.value, self.omega.value, self.Omega.value])

  def stats(self, returned=False):
    ''' 
    Prints a summary of the current stats of the Star.
    '''
    s = f"<{self.__module__}.{type(self).__name__} object at {hex(id(self))}, "
    s += f"m= {self.mass:1.4g}, "
    s += f"b= {self.impact_parameter:1.4g}, "
    s += f"v= {self.velocity:1.4g}, "
    s += f"inc= {self.inc:1.4g}, "
    s += f"omega= {self.omega:1.4g}, "
    s += f"Omega= {self.Omega:1.4g}>"
    if returned: return s
    else: print(s)
  
  def __str__(self):
    return self.stats(returned=True)
  
  def __repr__(self):
    return self.stats(returned=True)
  
  def __len__(self):
    return -1


class Stars(MutableMapping):
  '''
    This class allows the user to access stars like a dictionary using the star's index.
    Allows for negative indices and slicing.
    The implementation uses astropy.Quantity and numpy ndarrays and only generates a airball.Star object when a single Star is requested.
  '''
  def __init__(self, **kwargs) -> None:
    try: 
      self.units = UnitSystem(kwargs['UNIT_SYSTEM'])
      del kwargs['UNIT_SYSTEM']
    except KeyError: self.units = UnitSystem()

    self._Nstars = 0
    for key in kwargs:
      try: 
        len(kwargs[key])
        if isList(kwargs[key]) and len(kwargs[key]) > self.N:
          self._Nstars = len(kwargs[key])
      except: pass
    if 'size' in kwargs and self.N != 0: raise OverspecifiedParametersException('If lists are given then size cannot be specified.')
    elif 'size' in kwargs: self._Nstars = int(kwargs['size'])
    elif self.N == 0: raise UnspecifiedParameterException('If no lists are given then size must be specified.')
    else: pass

    
    try:
      # Check to see if was key is given.
      value = kwargs['m']
      # Check if length matches other key values.
      if len(value) != self.N: raise ListLengthException()
      # Length of value matches other key values, check if value is a list.
      elif isinstance(value, list):
        # Value is a list, try to turn list of Quantities into a ndarray Quantity.
        try: self._m = _numpy.array([v.to(self.units.units['mass']).value for v in value]) * self.units.units['mass']
        # Value was not a list of Quantities, turn list into ndarray and make a Quantity.
        except: self._m = _numpy.array(value) * self.units.units['mass']
      # Value was not a list, check to see if value is an ndarray.
      elif isinstance(value, _numpy.ndarray):
        # Assume ndarray is a Quantity and try to convert ndarray into given units.
        try: self._m = value.to(self.units.units['mass'])
        # ndarray is not a Quantity so turn ndarray into a Quantity.
        except: self._m = value * self.units.units['mass']
      # Value implements __len__, but is not a list or ndarray.
      else: raise IncompatibleListException()
    # This key is necessary and must be specified, raise and Exception.
    except KeyError: raise UnspecifiedParameterException('Mass, m, must be specified.')
    # Value is not a list, so assume it is an int or float and generate an ndarray of the given value.
    except TypeError: 
      value = value.to(self.units.units['mass']) if isQuantity(value) else value * self.units.units['mass']
      self._m = _numpy.ones(self.N) * value
    # Catch any additional Exceptions.
    except Exception as err: raise err


    try:
      # Check to see if was key is given.
      value = kwargs['b']
      # Check if length matches other key values.
      if len(value) != self.N: raise ListLengthException()
      # Length of value matches other key values, check if value is a list.
      elif isinstance(value, list):
        # Value is a list, try to turn list of Quantities into a ndarray Quantity.
        try: self._b = _numpy.array([v.to(self.units.units['length']).value for v in value]) * self.units.units['length']
        # Value was not a list of Quantities, turn list into ndarray and make a Quantity.
        except: self._b = _numpy.array(value) * self.units.units['length']
      # Value was not a list, check to see if value is an ndarray.
      elif isinstance(value, _numpy.ndarray):
        # Assume ndarray is a Quantity and try to convert ndarray into given units.
        try: self._b = value.to(self.units.units['length'])
        # ndarray is not a Quantity so turn ndarray into a Quantity.
        except: self._b = value * self.units.units['length']
      # Value implements __len__, but is not a list or ndarray.
      else: raise IncompatibleListException()
    # This key is necessary and must be specified, raise and Exception.
    except KeyError: raise UnspecifiedParameterException('Impact Parameter, b, must be specified.')
    # Value is not a list, so assume it is an int or float and generate an ndarray of the given value.
    except TypeError: 
      value = value.to(self.units.units['length']) if isQuantity(value) else value * self.units.units['length']
      self._b = _numpy.ones(self.N) * value
    # Catch any additional Exceptions.
    except Exception as err: raise err

    try:
      # Check to see if was key is given.
      value = kwargs['v']
      # Check if length matches other key values.
      if len(value) != self.N: raise ListLengthException()
      # Length of value matches other key values, check if value is a list.
      elif isinstance(value, list):
        # Value is a list, try to turn list of Quantities into a ndarray Quantity.
        try: self._v = _numpy.array([v.to(self.units.units['velocity']).value for v in value]) * self.units.units['velocity']
        # Value was not a list of Quantities, turn list into ndarray and make a Quantity.
        except: self._v = _numpy.array(value) * self.units.units['velocity']
      # Value was not a list, check to see if value is an ndarray.
      elif isinstance(value, _numpy.ndarray):
        # Assume ndarray is a Quantity and try to convert ndarray into given units.
        try: self._v = value.to(self.units.units['velocity'])
        # ndarray is not a Quantity so turn ndarray into a Quantity.
        except: self._v = value * self.units.units['velocity']
      # Value implements __len__, but is not a list or ndarray.
      else: raise IncompatibleListException()
    # This key is necessary and must be specified, raise and Exception.
    except KeyError: raise UnspecifiedParameterException('Velocity, v, must be specified.')
    # Value is not a list, so assume it is an int or float and generate an ndarray of the given value.
    except TypeError: 
      value = value.to(self.units.units['velocity']) if isQuantity(value) else value * self.units.units['velocity']
      self._v = _numpy.ones(self.N) * value
    # Catch any additional Exceptions.
    except Exception as err: raise err

    try:
      # Check to see if was key is given.
      value = kwargs['inc']
        # Check to see if value for key is string.
      if isinstance(value, str):
        # Value is a string, check to see if value for key is valid.
        if value != 'uniform': raise InvalidValueForKeyException()
        # Value 'uniform' for key is valid, now generate an array of values for key.
        self._inc = (2.0*_numpy.pi * _uniform.rvs(size=self.N) - _numpy.pi) * self.units.units['angle']
      # Value is not a string, check if length matches other key values.
      elif len(value) != self.N: raise ListLengthException()
      # Length of value matches other key values, check if value is a list.
      elif isinstance(value, list):
        # Value is a list, try to turn list of Quantities into a ndarray Quantity.
        try: self._inc = _numpy.array([v.to(self.units.units['angle']).value for v in value]) * self.units.units['angle']
        # Value was not a list of Quantities, turn list into ndarray and make a Quantity.
        except: self._inc = _numpy.array(value) * self.units.units['angle']
      # Value was not a list, check to see if value is an ndarray.
      elif isinstance(value, _numpy.ndarray):
        # Assume ndarray is a Quantity and try to convert ndarray into given units.
        try: self._inc = value.to(self.units.units['angle'])
        # ndarray is not a Quantity so turn ndarray into a Quantity.
        except: self._inc = value * self.units.units['angle']
      # Value implements __len__, but is not a list or ndarray.
      else: raise IncompatibleListException()
    # Key does not exist, assume the user wants an array of values to automatically be generated.
    except KeyError: self._inc = (2.0*_numpy.pi * _uniform.rvs(size=self.N) - _numpy.pi) * self.units.units['angle']
    # Value is not a list, so assume it is an int or float and generate an ndarray of the given value.
    except TypeError: 
      value = value.to(self.units.units['angle']) if isQuantity(value) else value * self.units.units['angle']
      self._inc = _numpy.ones(self.N) * value
    # Catch any additional Exceptions.
    except Exception as err: raise err

    try:
      # Check to see if was key is given.
      value = kwargs['omega']
        # Check to see if value for key is string.
      if isinstance(value, str):
        # Value is a string, check to see if value for key is valid.
        if value != 'uniform': raise InvalidValueForKeyException()
        # Value 'uniform' for key is valid, now generate an array of values for key.
        self._omega = (2.0*_numpy.pi * _uniform.rvs(size=self.N) - _numpy.pi) * self.units.units['angle']
      # Value is not a string, check if length matches other key values.
      elif len(value) != self.N: raise ListLengthException()
      # Length of value matches other key values, check if value is a list.
      elif isinstance(value, list):
        # Value is a list, try to turn list of Quantities into a ndarray Quantity.
        try: self._omega = _numpy.array([v.to(self.units.units['angle']).value for v in value]) * self.units.units['angle']
        # Value was not a list of Quantities, turn list into ndarray and make a Quantity.
        except: self._omega = _numpy.array(value) * self.units.units['angle']
      # Value was not a list, check to see if value is an ndarray.
      elif isinstance(value, _numpy.ndarray):
        # Assume ndarray is a Quantity and try to convert ndarray into given units.
        try: self._omega = value.to(self.units.units['angle'])
        # ndarray is not a Quantity so turn ndarray into a Quantity.
        except: self._omega = value * self.units.units['angle']
      # Value implements __len__, but is not a list or ndarray.
      else: raise IncompatibleListException()
    # Key does not exist, assume the user wants an array of values to automatically be generated.
    except KeyError: self._omega = (2.0*_numpy.pi * _uniform.rvs(size=self.N) - _numpy.pi) * self.units.units['angle']
    # Value is not a list, so assume it is an int or float and generate an ndarray of the given value.
    except TypeError: 
      value = value.to(self.units.units['angle']) if isQuantity(value) else value * self.units.units['angle']
      self._omega = _numpy.ones(self.N) * value
    # Catch any additional Exceptions.
    except Exception as err: raise err

    try:
      # Check to see if was key is given.
      value = kwargs['Omega']
        # Check to see if value for key is string.
      if isinstance(value, str):
        # Value is a string, check to see if value for key is valid.
        if value != 'uniform': raise InvalidValueForKeyException()
        # Value 'uniform' for key is valid, now generate an array of values for key.
        self._Omega = (2.0*_numpy.pi * _uniform.rvs(size=self.N) - _numpy.pi) * self.units.units['angle']
      # Value is not a string, check if length matches other key values.
      elif len(value) != self.N: raise ListLengthException()
      # Length of value matches other key values, check if value is a list.
      elif isinstance(value, list):
        # Value is a list, try to turn list of Quantities into a ndarray Quantity.
        try: self._Omega = _numpy.array([v.to(self.units.units['angle']).value for v in value]) * self.units.units['angle']
        # Value was not a list of Quantities, turn list into ndarray and make a Quantity.
        except: self._Omega = _numpy.array(value) * self.units.units['angle']
      # Value was not a list, check to see if value is an ndarray.
      elif isinstance(value, _numpy.ndarray):
        # Assume ndarray is a Quantity and try to convert ndarray into given units.
        try: self._Omega = value.to(self.units.units['angle'])
        # ndarray is not a Quantity so turn ndarray into a Quantity.
        except: self._Omega = value * self.units.units['angle']
      # Value implements __len__, but is not a list or ndarray.
      else: raise IncompatibleListException()
    # Key does not exist, assume the user wants an array of values to automatically be generated.
    except KeyError: self._Omega = (2.0*_numpy.pi * _uniform.rvs(size=self.N) - _numpy.pi) * self.units.units['angle']
    # Value is not a list, so assume it is an int or float and generate an ndarray of the given value.
    except TypeError: 
      value = value.to(self.units.units['angle']) if isQuantity(value) else value * self.units.units['angle']
      self._Omega = _numpy.ones(self.N) * value
    # Catch any additional Exceptions.
    except Exception as err: raise err
  
  @property
  def N(self):
    return self._Nstars
  
  @property
  def median_mass(self):
    return _numpy.median([mass.value for mass in self.m]) * (self._stars[0].m).unit
  
  @property
  def mean_mass(self):
    return _numpy.mean([mass.value for mass in self.m]) * (self._stars[0].m).unit
  
  def __getitem__(self, key):
    int_types = int, _numpy.integer
    
    if isinstance(key, slice):
        if key == slice(None, None, None): return self
        else: return Stars(m=self.m[key], b=self.b[key], v=self.v[key], inc=self.inc[key], omega=self.omega[key], Omega=self.Omega[key], UNIT_SYSTEM=self.units.UNIT_SYSTEM)

    if isinstance(key, int_types):
        # accept negative indices
        if key < 0: key += self.N
        if key < 0 or key >= self.N: raise AttributeError(f'Index {key} used to access stars out of range.')
        i = int(key)
        return Star(m=self.m[i], b=self.b[i], v=self.v[i], inc=self.inc[i], omega=self.omega[i], Omega=self.Omega[i], UNIT_SYSTEM=self.units.UNIT_SYSTEM)
    
    raise StarInvalidKeyException() 

  def __setitem__(self, key, value):
    star_type = Star, Stars
    if isinstance(value, star_type):
      self.m[key], self.b[key], self.v[key], self.inc[key], self.omega[key], self.Omega[key] = value.params
    else: raise InvalidStarException()
  
  def __delitem__(self, key):
    raise ValueError('Cannot delete Star elements from Stars array.')

  def __iter__(self):
    for i in range(self.N):
      yield Star(m=self.m[i], b=self.b[i], v=self.v[i], inc=self.inc[i], omega=self.omega[i], Omega=self.Omega[i], UNIT_SYSTEM=self.units.UNIT_SYSTEM)

  def __len__(self):
    return self.N
  
  def sortby(self, key):
    inds = _numpy.arange(self.N)
    if key == 'm': inds = _numpy.argsort(self.m)
    elif key == 'b': inds = _numpy.argsort(self.b)
    elif key == 'v': inds = _numpy.argsort(self.v)
    elif key == 'inc': inds = _numpy.argsort(self.inc)
    elif key == 'omega': inds = _numpy.argsort(self.omega)
    elif key == 'Omega': inds = _numpy.argsort(self.Omega)
    else: raise InvalidValueForKeyException()

    self.m[:] = self.m[inds]
    self.b[:] = self.b[inds]
    self.v[:] = self.v[inds]
    self.inc[:] = self.inc[inds]
    self.omega[:] = self.omega[inds]
    self.Omega[:] = self.Omega[inds]

  @property
  def m(self):
    return self._m

  @property
  def mass(self):
    return self.m
  
  @property
  def b(self):
    return self._b

  @property
  def impact_parameter(self):
    return self.b

  @property
  def v(self):
    return self._v

  @property
  def velocity(self):
    return self.v

  @property
  def inc(self):
    return self._inc

  @property
  def inclination(self):
    return self.inc

  @property
  def argument_periastron(self):
    return self.omega

  @property
  def omega(self):
    return self._omega

  @property
  def Omega(self):
    return self._Omega

  @property
  def longitude_ascending_node(self):
    return self.Omega

  @property
  def params(self):
    return [self.m, self.b, self.v, self.inc, self.omega, self.Omega]
  
  @property
  def param_values(self):
    return _numpy.array([self.m, self.b, self.v, self.inc, self.omega, self.Omega])


################################
###### Custom Exceptions #######
################################

class InvalidStarException(Exception):
  def __init__(self): super().__init__('Object is not a valid airball.Star object.')

class StarInvalidKeyException(Exception):
  def __init__(self): super().__init__('Invalid key for Stars.')

class UnspecifiedParameterException(Exception):
  def __init__(self, message): super().__init__(message)

class OverspecifiedParametersException(Exception):
  def __init__(self, message): super().__init__(message)

class ListLengthException(Exception):
  def __init__(self): super().__init__('List arguments must be same length.')

class IncompatibleListException(Exception):
  def __init__(self): super().__init__('The given list type is not compatible. Please use a Python list or an ndarray.')

class InvalidValueForKeyException(Exception):
  def __init__(self): super().__init__('Invalid value for key.')


import rebound as _rebound
import numpy as _numpy
import joblib as _joblib

from . import units
from .tools import *

############################################################
#################### Flyby Functions #######################
############################################################

def flyby_particle(sim, star_mass=1, star_b=100, star_v=None,  star_e=None, star_omega='uniform', star_Omega='uniform', star_inc='uniform', star_rmax=2.5e5):
    '''
        Return a REBOUND Particle for a flyby star given a REBOUND Simulation and flyby parameters.
        
        Parameters
        ----------
        sim : the REBOUND Simulation (star and planets) that will experience the flyby star
        star_mass : the mass of the flyby star in units of Msun
        star_b : impact parameter of the flyby star in units of AU
        star_v : the relative velocity at infinity between the central star and the flyby star (hyperbolic excess velocity) in units of km/s. Only specify star_v OR star_e, not both.
        star_e : the eccentricity of the flyby star (e > 1). Only specify star_e OR star_v, not both.
        star_omega : the argument of periapsis of the flyby star
        star_Omega : the longitude of the ascending node of the flyby star
        star_inc : the inclination of the flyby star
        star_rmax : the starting distance of the flyby star in units of AU
    '''

    #################################################
    ## Calculation of Flyby Star Initial Conditions ## 
    #################################################

    # Calculate the orbital elements of the flyby star.
    rmax = star_rmax # This is the starting distance of the flyby star in AU
    e = determine_eccentricity(sim, star_mass, star_b, star_v=star_v, star_e=star_e)
    a = -star_b/_numpy.sqrt(e**2. - 1.) # Compute the semi-major axis of the flyby star
    l = -a*(e*e-1.) # Compute the semi-latus rectum of the hyperbolic orbit (-a because the semi-major axis is negative)
    f = _numpy.arccos((l/rmax-1.)/e) # Compute the true anomaly

    #################################################

    return _rebound.Particle(sim, m=star_mass, a=a, e=e, f=-f, omega=star_omega, Omega=star_Omega, inc=star_inc, hash='flybystar')



# def flyby(sim, star=None, m=0.3, b=1000, v=40,  e=None, inc='uniform', omega='uniform', Omega='uniform', rmax=2.5e5, hybrid=True, crossOverFactor=30, overwrite=False):
def flyby(sim, star, rmax=4e5, hybrid=True, crossOverFactor=30, overwrite=False):
    '''
        Simulate a stellar flyby to a REBOUND simulation.
        
        Because REBOUND Simulations are C structs underneath the Python, this function can pass the simulation by reference.
        This can be done by specifying overwrite=True. Meaning, any changes made inside this function to the REBOUND simulation are permanent.
        This function assumes that you are using a WHFAST integrator with REBOUND.
        
        Parameters
        ----------
        sim : the REBOUND Simulation (star and planets) that will experience the flyby star
        m : the mass of the flyby star in units of Msun
        b : impact parameter of the flyby star in units of AU
        v : the relative velocity at infinity between the central star and the flyby star (hyperbolic excess velocity) in units of km/s. Only specify v OR e, not both.
        e : the eccentricity of the flyby star (e > 1). Only specify e OR v, not both.
        inc : the inclination of the flyby star
        omega : the argument of periapsis of the flyby star
        Omega : the longitude of the ascending node of the flyby star
        rmax : the starting distance of the flyby star in units of AU
        hybrid: True/False, use IAS15 (instead of WHFast) for the closest approach if b < planet_a * crossOverFactor
        crossOverFactor: the value for when to switch integrators if hybrid=True
        overwrite: determines whether or not to return a copy of sim (overwrite=False) or integrate using the original sim (overwrite=True)
    '''

    # Do not overwrite given sim.
    if not overwrite: sim = sim.copy()
    sim_units = rebound_units(sim)
    
    # Extract the star's characteristics.
    m, b, v, omega, Omega, inc = star.params

    ##################################################
    ## Calculation of Flyby Star Initial Conditions ## 
    ##################################################

    # Calculate the orbital elements of the flyby star.
    e = determine_eccentricity(sim, m, b, star_v=v, star_e=None) # TODO: Allow user to define e instead of v.
    a = -b/_numpy.sqrt(e**2. - 1.) # Compute the semi-major axis of the flyby star
    l = -a*(e*e-1.) # Compute the semi-latus rectum of the hyperbolic orbit (-a because the semi-major axis is negative)

    rmax = rmax.to(units.au) if isQuantity(rmax) else rmax * units.au
    f = _numpy.arccos((l/rmax-1.)/e) # Compute the true anomaly

    #################################################
    
    # Add the flyby star to the simulation. 
    rot = _rebound.Rotation.to_new_axes(newz=sim.angular_momentum())
    sim.move_to_hel() # Move the system into the heliocentric, invariable plane frame of reference.
    sim.rotate(rot)
    sim.add(m=m.value, a=a.value, e=e.value, f=-f.value, omega=omega.value, Omega=Omega.value, inc=inc.value, hash='flybystar')
    sim.ri_whfast.recalculate_coordinates_this_timestep = 1 # Because a new particle was added, we need to tell REBOUND to recalculate the coordinates.
    sim.rotate(rot.inverse())
    sim.move_to_com() # Move the system back into the centre of mass, original frame for integrating.

    tperi = sim.particles['flybystar'].T - sim.t # Compute the time to periapsis for the flyby star from the current time.
    
    # Try to integrate the flyby. Start at the current time and go to twice the time to periapsis.
    # try:
    if hybrid:
        rCrossOver = crossOverFactor * sim.particles[1].a * sim_units['length'] # This is distance to switch integrators
        
        if b < rCrossOver:
            a = -b/_numpy.sqrt(e**2. - 1.) # Compute the semi-major axis of the flyby star
            l = -a*(e*e-1.) # Compute the semi-latus rectum of the hyperbolic orbit (-a because the semi-major axis is negative)
            f = _numpy.arccos((l/rCrossOver-1.)/e) # Compute the true anomaly

            G = (sim.G * sim_units['length']**3 / sim_units['mass'] / sim_units['time']**2)
            mu = G * (_numpy.sum([p.m for p in sim.particles]) * sim_units['mass'] + m)

            # Calculate half of the integration time for the flyby star.
            with units.set_enabled_equivalencies(units.dimensionless_angles()):
                E = _numpy.arccosh((_numpy.cos(f)+e)/(1.+e*_numpy.cos(f))) # Compute the eccentric anomaly
                M = e * _numpy.sinh(E)-E # Compute the mean anomaly

            tIAS15 = M/_numpy.sqrt(mu/(-a*a*a)) # Compute the time to periapsis (-a because the semi-major axis is negative)

            t1 = sim.t + tperi - tIAS15.value
            t2 = sim.t + tperi
            t3 = sim.t + tperi + tIAS15.value
            t4 = sim.t + 2*tperi

            dt = sim.dt
            dt_frac = sim.dt/sim.particles[1].P
            # print(f'\n::Initial::\ndt: {sim.dt:6.2f}\na: {sim.particles[1].a:6.2f}\ne: {sim.particles[1].e:6.2f}')

            sim.integrate(t1, exact_finish_time=0)
            sim.ri_whfast.recalculate_coordinates_this_timestep = 1
            sim.integrator_synchronize()

            sim.integrator = 'ias15'
            sim.gravity = 'basic'
            sim.integrate(t2)

            # de = airball.energy_change_close_encounters_sim(sim)
            sim.move_to_com()

            sim.integrate(t3)
            
            sim.integrator = 'whfast'
            sim.ri_whfast.safe_mode = 0
            sim.ri_whfast.recalculate_coordinates_this_timestep = 1
            sim.integrator_synchronize()
            if sim.particles[1].P > 0: sim.dt = dt_frac*sim.particles[1].P
            else: sim.dt = dt
            
            sim.integrate(t4, exact_finish_time=0)
        else:
            t1 = sim.t + tperi
            t2 = sim.t + 2*tperi

            sim.integrate(t1, exact_finish_time=0)
            sim.ri_whfast.recalculate_coordinates_this_timestep = 1
            sim.integrator_synchronize()
            # de = airball.energy_change_close_encounters_sim(sim)
            sim.move_to_com()
            sim.integrate(t2, exact_finish_time=0)
            sim.ri_whfast.recalculate_coordinates_this_timestep = 1
            sim.integrator_synchronize()
    else:
        t1 = sim.t + tperi
        t2 = sim.t + 2*tperi
        sim.integrate(t1, exact_finish_time=0)
        sim.ri_whfast.recalculate_coordinates_this_timestep = 1
        sim.integrator_synchronize()
        # de = airball.energy_change_close_encounters_sim(sim)
        sim.move_to_com()
        sim.integrate(t2, exact_finish_time=0)
        sim.ri_whfast.recalculate_coordinates_this_timestep = 1
        sim.integrator_synchronize()
    
    # Remove the flyby star. 
    sim.remove(hash='flybystar')
    sim.ri_whfast.recalculate_coordinates_this_timestep = 1 # Because a particle was removed, we need to tell REBOUND to recalculate the coordinates and to synchronize.
    sim.integrator_synchronize()
    sim.move_to_com() # Readjust the system back into the centre of mass/momentum frame for integrating.
    
    return sim
    # except Exception as err:
    #     print(err)
    #     return sim


def flybys(sims, stars, **kwargs):
    '''
        Run serial flybys in parallel.
    '''
    Nruns = 0
    try: 
        Nruns = len(sims)
        if Nruns != len(stars): raise Exception('Sims and stars are unequal lengths')
    except: 
        Nruns = len(stars)
        sims = [sims.copy() for _ in range(Nruns)]
    
    try: inds = kwargs['inds']
    except KeyError: inds = _numpy.arange(Nruns)

    try: 
        rmax = kwargs['rmax']
        if not isList(rmax): rmax = Nruns * [rmax]
        elif len(rmax) != Nruns: raise Exception('List arguments must be same length.')
    except KeyError: rmax = Nruns * [4e5]

    try: 
        hybrid = kwargs['hybrid']
        if not isList(hybrid): hybrid = Nruns * [hybrid]
        elif len(hybrid) != Nruns: raise Exception('List arguments must be same length.')
    except KeyError: hybrid = Nruns * [True]

    try: 
        crossOverFactor = kwargs['crossOverFactor']
        if not isList(crossOverFactor): crossOverFactor = Nruns * [crossOverFactor]
        elif len(crossOverFactor) != Nruns: raise Exception('List arguments must be same length.')
    except KeyError: crossOverFactor = Nruns * [30]

    try: overwrite = kwargs['overwrite']
    except KeyError: overwrite = False

    try: n_jobs = kwargs['n_jobs']
    except KeyError: n_jobs = -1

    try: verbose = kwargs['verbose']
    except KeyError: verbose = 0

    sim_results = _joblib.Parallel(n_jobs=n_jobs, verbose=verbose, require='sharedmem')(
    _joblib.delayed(flyby)(
        sim=sims[int(i)], star=stars[i], rmax=rmax[i], hybrid=hybrid[i], crossOverFactor=crossOverFactor[i], overwrite=overwrite) 
    for i in inds)
    
    return sim_results




import rebound as _rebound
import numpy as _numpy
import json as _json
from ctypes import c_double, pointer, c_int, byref

from .tools import *

def notNone(a):
    """
    Returns True if array a contains at least one element that is not None. Returns False otherwise.
    """
    return a.count(None) != len(a)

class Particle():
    """
        AIRBALL Particle object.
        
        The Particle class is a convenient way to generate flyby objects used in REBOUND simulations. Each exostar is randomly generated (but are always seeded) and is always generated in heliocentric coordinates.

        To create a random Particle object simply type
        >>> ex = Particle()

        The specific parameters of the object can be defined using keyword arguments. All the current parameters that can be specified are in default REBOUND units (AU, Msun, Yr2Pi), and they are:
            exomass             # The mass of the Particle object (default value is 1 Msun).
            primary_mass        # The combined mass of the system the Particle will flyby (default value is 1 Msun).
            seed                # The random seed used in specifying the location and direction of the flyby (default value is randomly selected and saved).
            b                   # The impact parameter (default value is 100 AU).
            b_min,b_max         # Two arguments that specify a range to uniformly select the impact parameter from.
            periastron          # If the impact parameter is not specified, the periastron of the Particle can be specified instead (default value is 100 AU).
            peri_min,peri_max   # If the impact parameter is not specified, the these two arguments that specify a range to uniformly select the periastron from.
            start_dist          # Similar to the REBOUND simulation exit_max_distance property. This specifies how far away the Particle starts in the simulation, effectively the distance at which the influence of the Particle becomes negligable (default value is 1000 AU).
            vinf                # The hyperbolic excess velocity, or the velocity that the Particle would have at infinity (default value is about 0.0335 AU/Yr2Pi or 1 km/s).
            vinf_units          # If the default hyperbolic excess velocity is changed, units of km/s can be specified for convenience (default value is "None", but "km/s" are available).
            from_file           # This is used for loading an Particle object from a binary file (see below).
        
        To save an Particle object you can use 
        >>> ex.save('filename.bin')

        And then to load the Particle object, simply use
        >>> ex = Exostar.load_from_file('filename.bin')
        
    """
    def __init__(self, exomass=None, primary_mass=None, seed=None, b=None, b_min=None, b_max=None, periastron=None, peri_min=None, peri_max=None, start_dist=None, vinf=None, vinf_units=None, from_file=None, vdir=None, rdir=None):
        if from_file is None:
            # Set up characteristics of the perturbing exostar in a heliocentric frame.
            #   Take the primary_mass to be the combined mass of the host star and all planets.
            if exomass is None:
                self._mass = 1.0
            else:
                self._mass = exomass
            if primary_mass is None:
                self._primary_mass = 1.0
            elif type(primary_mass) is _rebound.simulation.Simulation:
                self._primary_mass = sum([p.m for p in primary_mass.particles])
            else:
                self._primary_mass = primary_mass

            # Set seed for random orientation.
            if seed is None:
                self._seed = int(_numpy.random.randint(0, 2**32 - 1))
            else:
                self._seed = int(seed)
            _numpy.random.seed(self._seed)

            # Set periastron or impact parameter.
            self._peri = None
            self._b = None
            peri_set = [periastron, peri_min, peri_max]
            b_set = [b, b_min, b_max]
            if notNone(peri_set) and notNone(b_set):
                raise AssertionError('You cannot specify the periastron and the impact parameter b.')
            if notNone(peri_set):
                if peri_min is None:
                    if peri_max is None:
                        self._peri = periastron
                    elif periastron is None:
                        raise ValueError('You cannot specify only a maximum periastron. Specify a range with peri_min and peri_max, or specify a specific periastron.')
                    else:
                        raise ValueError('You cannot a specific periastron and a maximum periastron. Specify a range with peri_min and peri_max, or specify a specific periastron.')
                elif peri_max is None:
                    if periastron is None:
                        raise ValueError('You cannot specify only a minimum periastron. Specify a range with peri_min and peri_max, or specify a specific periastron.')
                    else:
                        raise ValueError('You cannot a specific periastron and a minimum periastron. Specify a range with peri_min and peri_max, or specify a specific periastron.')
                elif periastron is None:
                    self._peri = _numpy.random.uniform(peri_min, peri_max)
                else:
                    raise ValueError('You cannot a specific periastron and a periastron range. Specify a range with peri_min and peri_max, or specify a specific periastron.')
            if notNone(b_set):
                if b_min is None:
                    if b_max is None:
                        self._b = b
                    elif b is None:
                        raise ValueError('You cannot specify only a maximum impact parameter. Specify a range with b_min and b_max, or specify a specific impact parameter b.')
                    else:
                        raise ValueError('You cannot a specific impact parameter and a minimum impact parameter. Specify a range with b_min and b_max, or specify a specific impact parameter.')
                elif b_max is None:
                    if b is None:
                        raise ValueError('You cannot specify only a minimum impact parameter. Specify a range with b_min and b_max, or specify a specific impact parameter.')
                    else:
                        raise ValueError('You cannot a specific impact parameter and a minimum impact parameter. Specify a range with b_min and b_max, or specify a specific impact parameter.')
                elif b is None:
                    self._b = _numpy.random.uniform(b_min, b_max)
                else:
                    raise ValueError('You cannot a specific impact parameter and an impact parameter range. Specify a range with b_min and b_max, or specify a specific impact parameter.')
            
            # Set distance between star and exostar.
            d = None
            if self._peri is None and self._b is not None:
                d = self._b
            elif self._b is None and self._peri is not None:
                d = self._peri
            else:
                self._b = 100.
                d = self._b
            
            # Randomly vary the orientation of the exostar at the periapse.
            if vdir is None:
                self._v = self._vec(3)
            else:
                self._v = _numpy.array(vdir)/_numpy.linalg.norm(vdir)
            if rdir is None:
                self._r = _numpy.zeros(3)
                self._r[0:2] = d * self._vec(2)
                self._r = _numpy.dot(self._r, self._R(self._v))
            else:
                self._r = d * _numpy.array(rdir)/_numpy.linalg.norm(rdir)

            # Set set the magnitudes of the position and velocity vectors.
            if vinf is not None:
                if vinf_units is None:
                    self._vinf = vinf
                elif vinf_units == 'km/s':
                    au = 149597870700. # meters in 1 AU
                    yr2pi = (86400.*365.25)/twopi # seconds in 1 Yr2Pi
                    self._vinf = (vinf*1000.)*yr2pi/au # In units of AU/Yr2Pi
                else:
                    raise ValueError('The only other units available are "km/s". Otherwise leave as "None".', )
            else: 
                self._vinf = 0.03357365989646265980 # 1.0 km/s in units of AU/Yr2Pi

            rmag = _numpy.linalg.norm(self._r)
            vmag = _numpy.sqrt( self._vinf**2 + 2. * (self._mass + self._primary_mass) / rmag)
            self._v = self._v * vmag

            # Set the exit distance (how far away before the effects of the exostar are negligable). 
            # TODO: Change the starting distance to be dependent on the mass of the exostar.
            if start_dist is None:
                if self._peri is None:
                    if self._b < 1000.:
                        self._start_dist = 1000.
                    else:
                        self._start_dist = 10.*self._b
                elif self._b is None:
                    if self._peri < 1000.:
                        self._start_dist = 1000.
                    else:
                        self._start_dist = 10.*self._peri
            else:
                self._start_dist = start_dist

            # If the impact parameter or the periastron are less than the starting distance, then move the Particle backwards to the starting distance.
            if self._b is None:
                if self._start_dist < self._peri:
                    raise AssertionError('The periastron cannot be larger than the starting distance.')
                
                # Solve Kepler's Equation for the time it takes to move to the starting position.
                # Assuming Natural Units: G = 1 (Msun = 1, AU = 1, Period = Yr2Pi)
                mu = self._primary_mass + self._mass
                a = -mu/self._vinf**2.
                e = 1. - (rmag/a)
                r = self._start_dist
                cosf = (a * (1 - e**2) - r)/(e*r)
                coshF = (e + cosf)/(1 + e*cosf)
                sinhF = _numpy.sqrt(coshF**2 - 1)
                self._start_time = -_numpy.sqrt(((-a)**3)/mu) * (e * sinhF - _numpy.arcsinh(sinhF) )

                # Solve Kepler's Equation for new position.
                _p = _rebound.Particle(m=self._mass, x=self._r[0], y=self._r[1], z=self._r[2], vx=self._v[0], vy=self._v[1], vz=self._v[2])
                _rebound.clibrebound.reb_whfast_kepler_solver(byref(_rebound.Simulation()), byref(_p), c_double(mu), c_int(0), c_double(self._start_time))

                # Extract the Exostar coordinates in the heliocentric frame.
                self._r = _p.xyz
                self._v = _p.vxyz
            elif self._peri is None:
                if self._start_dist < self._b:
                    raise AssertionError('The impact parameter cannot be larger than the starting distance.')
                alpha = _numpy.sqrt(self._start_dist**2 - _numpy.linalg.norm(self._r)**2)
                v = self._v/_numpy.linalg.norm(self._v)
                self._r = (self._r - alpha*v).tolist()
                self._v = self._v.tolist()
                sim = _rebound.Simulation()
                sim.add(m=self._primary_mass)
                sim.add(self.particle)
                self._start_time = -sim.calculate_orbits()[0].T
            else:
                self._start_time = 0.0
        else:
            self.__dict__.update(from_file)
    # End of Constructor

    def save(self, filename):
        dic_copy = self.__dict__.copy()
        _json.dump(dic_copy, open(filename, "w"), ensure_ascii=False)

    @classmethod
    def load_from_file(cls, filename):
        dic = _json.load(open(filename, "r"))
        return cls(from_file=dic)
    
    @classmethod
    def from_dict(cls, dic):
        return cls(from_file=dic)

    @property
    def particle(self):
        '''Returns a REBOUND particle of the exostar in heliocentric coordinates.'''
        return _rebound.Particle(m=self._mass, x=self._r[0], y=self._r[1], z=self._r[2], vx=self._v[0], vy=self._v[1], vz=self._v[2])
    
    @property
    def primary_mass(self):
        return self._primary_mass

    @property
    def start_time(self):
        '''Returns how long it will take the exostar to travel from it's current location to periastron.'''
        return self._start_time

    @property
    def m(self):
        return self._mass

    @property
    def x(self):
        return self._r[0]

    @property
    def y(self):
        return self._r[1]

    @property
    def z(self):
        return self._r[2]

    @property
    def xyz(self):
        '''Returns the position vector.'''
        return self._r

    @property
    def vx(self):
        return self._v[0]

    @property
    def vy(self):
        return self._v[1]

    @property
    def vz(self):
        return self._v[2]

    @property
    def vxyz(self):
        '''Returns the velocity vector.'''
        return self._v

    @property
    def r(self):
        '''Returns the magnitude of the position vector.'''
        return _numpy.linalg.norm(self._r)

    @property
    def v(self):
        '''Returns the magnitude of the velocity vector.'''
        return _numpy.linalg.norm(self._v)

    @property
    def vinf(self):
        '''Returns the magnitude of the hyperbolic excess velocity, or the velocity that the Particle would have at infinity.'''
        return _numpy.linalg.norm(self._vinf)

    @property
    def pe(self):
        if self._peri is None:
            sim = _rebound.Simulation()
            sim.add(m=self._primary_mass)
            sim.add(self.particle)
            self._peri = sim.particles[1].a*(1 - sim.particles[1].e)
        return self._peri

    @property
    def b(self):
        if self._b is None:
            sim = _rebound.Simulation()
            sim.add(m=self._primary_mass)
            sim.add(self.particle)
            self._b = -sim.particles[1].a * _numpy.sqrt(sim.particles[1].e**2. - 1.)
            # raise ValueError('The impact parameter is not defined.')
        return self._b
    
    @property
    def seed(self):
        return self._seed

    def _vec(self, d):
        '''
            Returns a normalized d-vector.
        '''
        r = 2.
        while r > 1.: 
            out = 2. * _numpy.random.rand(d) - 1.
            r = _numpy.linalg.norm(out)
        return out/r

    def _R(self, v):
        '''
            Given a 3-vector "v", returns a matrix used for rotating other vectors to be perpendicular to "v".
        '''
        c = lambda x: _numpy.cos(x)
        s = lambda x: _numpy.sin(x)
        T = float(_numpy.arcsin(v[1]))
        P = float(_numpy.arcsin(v[0]/c(T)))
        rotation = _numpy.array([[c(P), 0, -s(P)], [-s(T)*s(P), c(T), -s(T)*c(P)], [c(T)*s(P), s(T), c(T)*c(P)]])
        flip = _numpy.array([[1, 0, 0], [0, 1, 0], [0, 0, _numpy.sign(v[2])]])
        return _numpy.dot(rotation,flip)



import rebound as _rebound
import numpy as _numpy
from scipy.stats import uniform as _uniform
from scipy.stats import maxwell as _maxwell
from scipy.optimize import fminbound as _fminbound

from . import units
from .imf import *
from .stars import *
from .analytic import *
from .tools import *

class StellarEnvironment:
  '''
    This is the AIRBALL StellarEnvironment class.
    It encapsulates the relevant data for a static stellar environment.

    # Example
    my_env = airball.StellarEnvironment(stellar_density=10, velocity_dispersion=20, lower_mass_limit=0.08, upper_mass_limit=8, name='My Environment')
    my_star = my_env.random_star()

    If a `maximum_impact_parameter` is not given, AIRBALL attempts to estimate a reasonable one.
    There are predefined subclasses for the LocalNeighborhood, a generic OpenCluster, a generic GlobularCluster, and the Milky Way center GalacticBulge and GalacticCore.
  '''
  def __init__(self, stellar_density, velocity_dispersion, lower_mass_limit, upper_mass_limit, mass_function=None, maximum_impact_parameter=None, name=None, UNIT_SYSTEM=[], object_name=None):

    # Check to see if an stars object unit is defined in the given UNIT_SYSTEM and if the user defined a different name for the objects.
    objectUnit = [this for this in UNIT_SYSTEM if this.is_equivalent(units.stars)]
    if objectUnit == [] and object_name is not None: UNIT_SYSTEM.append(units.def_unit(object_name, units.stars))
    self.units = UnitSystem(UNIT_SYSTEM)

    self.density = stellar_density
    self.velocity_dispersion = velocity_dispersion
    self.maximum_impact_parameter = maximum_impact_parameter

    self._upper_mass_limit = upper_mass_limit.to(self.units.units['mass']) if isQuantity(upper_mass_limit) else upper_mass_limit * self.units.units['mass']
    self._lower_mass_limit = lower_mass_limit.to(self.units.units['mass']) if isQuantity(lower_mass_limit) else lower_mass_limit * self.units.units['mass']
    self._IMF = IMF(min_mass=self._lower_mass_limit, max_mass=self._upper_mass_limit, mass_function=mass_function, unit=self.units.units['mass'])
    self._median_mass = self.IMF.median_mass

    self.name = name if name is not None else 'Stellar Environment'

  def random_star(self, maximum_impact_parameter=None, include_orientation=True, size=1):
    '''
      Computes a random star from a stellar environment.
      Returns: airball.Star() or airball.Stars() if size > 1.
    '''
    size = int(size)

    v = _maxwell.rvs(scale=maxwell_boltzmann_scale_from_dispersion(self.velocity_dispersion), size=size) # Relative velocity of the star at infinity.

    max_impact = maximum_impact_parameter if maximum_impact_parameter is not None else self.maximum_impact_parameter
    b = max_impact * _numpy.sqrt(_uniform.rvs(size=size)) # Impact parameter of the star.
    
    m = self.IMF.random_mass(size=size) # Mass of the star.
    
    zeros = _numpy.zeros(size)
    inc = 2.0*_numpy.pi * _uniform.rvs(size=size) - _numpy.pi if include_orientation else zeros
    ω = 2.0*_numpy.pi * _uniform.rvs(size=size) - _numpy.pi if include_orientation else zeros
    Ω = 2.0*_numpy.pi * _uniform.rvs(size=size) - _numpy.pi if include_orientation else zeros

    if size > 1: return Stars(m=m, b=b, v=v, inc=inc, omega=ω, Omega=Ω)
    else: return Star(m, b[0], v[0], inc[0], ω[0], Ω[0])

  def stats(self):
    ''' 
    Prints a summary of the current stats of the Stellar Environment.
    '''
    s = self.name
    s += "\n------------------------------------------\n"
    s += "Stellar Density:     {0:12.4g} \n".format(self.density)
    s += "Velocity Scale:      {0:12.4g} \n".format(self.velocity_dispersion)
    s += "Mass Range:            {0:6.4g} - {1:1.4g}\n".format(self.lower_mass_limit.value, self.upper_mass_limit)
    s += "Median Mass:         {0:12.4g} \n".format(self.median_mass)
    s += "Max Impact Param:    {0:12.4g} \n".format(self.maximum_impact_parameter)
    s += "Encounter Rate:      {0:12.4g} \n".format(self.encounter_rate)
    s += "------------------------------------------"
    print(s)

  @property
  def object_unit(self):
    return self.units.units['object']
  
  @property
  def UNIT_SYSTEM(self):
      return self._UNIT_SYSTEM

  @UNIT_SYSTEM.setter
  def UNIT_SYSTEM(self, UNIT_SYSTEM):
    self.units.UNIT_SYSTEM = UNIT_SYSTEM

  @property
  def median_mass(self):
    '''
      The median mass of the environment's IMF
    '''
    return self.IMF.median_mass.to(self.units.units['mass'])

  @property
  def maximum_impact_parameter(self):
    '''
      Compute an estimate for the largest impact parameter to affect a Sun-Neptune system.
    '''
    # TODO: Update with UNITS
    if self._maximum_impact_parameter is None:
      sim = _rebound.Simulation()
      sim.add(m=1.0)
      sim.add(m=5.2e-05, a=30.2, e=0.013)
      _f = lambda b: _numpy.log10(_numpy.abs(1e-16 - _numpy.abs(relative_energy_change(sim, Star(self.upper_mass_limit, b, _numpy.sqrt(2.0)*maxwell_boltzmann_mean_from_dispersion(self.velocity_dispersion))))))
      bs = _numpy.logspace(1, 6, 1000)
      b0 = bs[_numpy.argmin(_f(bs))]
      self._maximum_impact_parameter = _fminbound(_f, b0/5, 5*b0) * units.au
    return self._maximum_impact_parameter.to(self.units.units['length'])
  
  @maximum_impact_parameter.setter
  def maximum_impact_parameter(self, value):
    if value is not None:
      self._maximum_impact_parameter = value.to(self.units.units['length']) if isQuantity(value) else value * self.units.units['length']

  @property
  def density(self):
    '''
      The number density of the environment.
      Default units: pc^{-3}.
    '''
    return self._density.to(self.units.units['density'])

  @density.setter
  def density(self, value):
    '''
      The number density of the environment.
      Default units: pc^{-3}.
    '''
    if isQuantity(value):
      if value.unit.is_equivalent(units.stars/units.m**3): self._density = value.to(self.units.units['density'])
      elif value.unit.is_equivalent(1/units.m**3): self._density = (value * self.units.units['object']).to(self.units.units['density'])
      else: AssertionError('The given density units are not compatible.')
    else: self._density = value * self.units.units['density']

  @property
  def velocity_dispersion(self):
    '''
      Return the velocity dispersion of the environment.
      Default units: km/s.
    '''
    return self._velocity.to(self.units.units['velocity'])
  
  @velocity_dispersion.setter
  def velocity_dispersion(self, value):
    '''
      The velocity dispersion of the environment.
      Default units: km/s.
    '''
    self._velocity = value.to(self.units.units['velocity']) if isQuantity(value) else value * self.units.units['velocity']
  
  @property
  def velocity_mean(self):
    '''
      Return the velocity dispersion of the environment.
      Default units: km/s.
    '''
    return maxwell_boltzmann_mean_from_dispersion(self.velocity_dispersion).to(self.units.units['velocity'])
  
  @property
  def velocity_mode(self):
    '''
      Return the velocity dispersion of the environment.
      Default units: km/s.
    '''
    return maxwell_boltzmann_mode_from_dispersion(self.velocity_dispersion).to(self.units.units['velocity'])
  
  @property
  def velocity_rms(self):
    '''
      Return the velocity dispersion of the environment.
      Default units: km/s.
    '''

    v = _maxwell.rvs(scale=maxwell_boltzmann_scale_from_dispersion(self.velocity_dispersion), size=int(1e6))
    return verify_unit(_numpy.sqrt(_numpy.mean(v**2)), self.units.units['velocity'])

  @property
  def lower_mass_limit(self):
    '''
      Return the lower mass limit of the IMF of the environment.
      Default units: solMass
    '''
    return self.IMF.min_mass.to(self.units.units['mass'])
  
  @lower_mass_limit.setter
  def lower_mass_limit(self, value):
    '''
      The lower mass limit of the IMF of the environment.
      Default units: solMass
    '''
    self.IMF.min_mass = value

  @property
  def upper_mass_limit(self):
    '''
      Return the lower mass limit of the IMF of the environment.
      Default units: solMass
    '''
    return self.IMF.max_mass.to(self.units.units['mass'])
  
  @upper_mass_limit.setter
  def upper_mass_limit(self, value):
    '''
      The lower mass limit of the IMF of the environment.
      Default units: solMass
    '''
    self.IMF.max_mass = value


  @property
  def IMF(self):
    '''
      Return the IMF of the environment.
    '''
    return self._IMF
  
  @IMF.setter
  def IMF(self, value):
    '''
      The IMF of the environment.
    '''
    if isinstance(value, IMF): self._IMF = IMF(value.min_mass, value.max_mass, value.imf, value.unit, value.number_samples, value.seed)
    else: raise AssertionError('Initial Mass Function (IMF) must be an airball.IMF object.')
  
  @property
  def encounter_rate(self):
    '''
        Compute the expected flyby encounter rate Γ = ⟨nσv⟩ for the stellar environment in units of flybys per year.
        The inverse of the encouter rate will give the average number of years until a flyby.

        n : stellar number density. Default units: pc^{-3}
        σ : interaction cross section. Default units: AU^2
        v : velocity dispersion. Default units: km/s

        The interaction cross section σ = πb^2 considers gravitational focussing b = q√[1 + (2GM)/(q v∞^2)] and considers
        - the median mass of the environment
        - the maximum impact parameter
        - the relative velocity at infinity derived from the velocity dispersion
    '''
    return encounter_rate(self._density, maxwell_boltzmann_mean_from_dispersion(self.velocity_dispersion), self._maximum_impact_parameter, self.median_mass).to(self.units.units['object']/self.units.units['time'])



class LocalNeighborhood(StellarEnvironment):
  '''
    This is a AIRBALL StellarEnvironment subclass for the Local Neighborhood.
    It encapsulates the relevant data for a static stellar environment representing the local neighborhood of the solar system.

    The stellar density is 0.14 pc^-3 defined by Bovy (2017).
    The velocity dispersion is 20 km/s, defined by Binnery & Tremaine (2008) v_rms ~50 km/s and Bailer-Jones et al. (2018) so that 90% of stars have v < 100 km/s with an encounter rate of ~20 stars/Myr within 1 pc.
    The mass limits is defined to between 0.08-8 solar masses using Chabrier (2003) for single stars when m < 1 and a power-law model from Bovy (2017) for stars m ≥ 1 to account for depleted stars due to stellar evolution.

    # Example
    my_local = airball.LocalNeighborhood()
    my_10stars = my_local.random_star(size=10)
    # returns a Stars object with the masses, impact parameters, velocities, and orientation of the 10 Star objects in a heliocentric model.
  '''
  short_name = 'Local'
  def local_mass_function(x):
    '''
      This defined using Chabrier (2003) for single stars when m < 1 and a power-law model from Bovy (2017) for stars m ≥ 1 to account for depleted stars due to stellar evolution.
    '''
    return chabrier_2003_single(1, 0.0567) * (x)**-4.7 if x > 1 else chabrier_2003_single(x, 0.0567)

  def __init__(self, stellar_density = 0.14 * units.stars/units.pc**3, velocity_dispersion = 20.8 * units.km/units.s, lower_mass_limit=0.08 * units.solMass, upper_mass_limit = 8 * units.solMass, maximum_impact_parameter=None, UNIT_SYSTEM=[], mass_function=local_mass_function):
    super().__init__(stellar_density=stellar_density, velocity_dispersion=velocity_dispersion, lower_mass_limit=lower_mass_limit, upper_mass_limit=upper_mass_limit, mass_function=mass_function, maximum_impact_parameter=maximum_impact_parameter, UNIT_SYSTEM=UNIT_SYSTEM, name = 'Local Neighborhood')
    self.maximum_impact_parameter = 10000 * units.au if maximum_impact_parameter is None else maximum_impact_parameter

class OpenCluster(StellarEnvironment):
  '''
    This is a AIRBALL StellarEnvironment subclass for a generic Open Cluster.
    It encapsulates the relevant data for a static stellar environment representing a generic open cluster.

    The stellar density is 100 pc^-3 informed by Adams (2010).
    The velocity scale is 1 km/s informed by Adams (2010) and Malmberg et al. (2011).
    The mass limit is defined to between 0.01-100 solar masses using Chabrier (2003) for single stars when m < 1 and Salpeter (1955) for stars m ≥ 1.

    # Example
    my_open = airball.OpenCluster()
    my_10stars = my_open.random_star(size=10)
    # returns a Stars object with the masses, impact parameters, velocities, and orientation of the 10 Star objects in a heliocentric model.
  '''
  short_name = 'Open'
  
  def __init__(self, maximum_impact_parameter=None, UNIT_SYSTEM=[]):
    super().__init__(stellar_density = 100 * units.stars * units.pc**-3, velocity_dispersion = 1 * units.km/units.s, lower_mass_limit=0.08 * units.solMass, upper_mass_limit = 100 * units.solMass, maximum_impact_parameter=maximum_impact_parameter, UNIT_SYSTEM=UNIT_SYSTEM, name = 'Open Cluster')
    self._maximum_impact_parameter = 1000 * units.au

class GlobularCluster(StellarEnvironment):
  short_name = 'Globular'
  
  def __init__(self, maximum_impact_parameter=None, UNIT_SYSTEM=[]):
    super().__init__(stellar_density = 1000 * units.stars * units.pc**-3, velocity_dispersion = 10 * units.km/units.s, lower_mass_limit=0.08 * units.solMass, upper_mass_limit = 1 * units.solMass, maximum_impact_parameter=maximum_impact_parameter, UNIT_SYSTEM=UNIT_SYSTEM, name = 'Globular Cluster')
    self._maximum_impact_parameter = 5000 * units.au

class GalacticBulge(StellarEnvironment):
  short_name = 'Bulge'
  
  def __init__(self, maximum_impact_parameter=None, UNIT_SYSTEM=[]):
    super().__init__(stellar_density = 50 * units.stars * units.pc**-3, velocity_dispersion = 120 * units.km/units.s, lower_mass_limit=0.08 * units.solMass, upper_mass_limit = 10 * units.solMass, maximum_impact_parameter=maximum_impact_parameter, UNIT_SYSTEM=UNIT_SYSTEM, name = 'Milky Way Bulge')
    self._maximum_impact_parameter = 50000 * units.au

class GalacticCore(StellarEnvironment):
  short_name = 'Core'
  
  def __init__(self, maximum_impact_parameter=None, UNIT_SYSTEM=[]):
    super().__init__(stellar_density = 10000 * units.stars * units.pc**-3, velocity_dispersion = 170 * units.km/units.s, lower_mass_limit=0.08 * units.solMass, upper_mass_limit = 10 * units.solMass, maximum_impact_parameter=maximum_impact_parameter, UNIT_SYSTEM=UNIT_SYSTEM, name = 'Milky Way Core')
    self._maximum_impact_parameter = 50000 * units.au




import numpy as _numpy
from scipy.integrate import quad as _quad
from . import units
from .tools import *


def chabrier_2003_single(x, A=0.158):
    """
    Chabrier 2003 IMF for single stars.
    This function calculates the probability density for a given mass value (x) based on the Chabrier 2003 IMF equation.

    Parameters:
    - x: Mass value (float).

    Returns:
    - Probability density at the given mass value.
    """
    return (A / x) * _numpy.exp(-((_numpy.log10(x) - _numpy.log10(0.079)) ** 2) / (2 * 0.69 ** 2))

def salpeter_1955(x, A):
    """
    Salpeter 1955 IMF for single stars.
    This function calculates the probability density for a given mass value (x) based on the Salpeter 1955 IMF equation.

    Parameters:
    - x: Mass value (float).

    Returns:
    - Probability density at the given mass value.
    """
    return A * x ** -2.3


class IMF():
  """
  Class representing an Initial Mass Function (IMF).
  It generates random masses based on a given mass function and provides various properties and methods for manipulating and analyzing the IMF.

  Parameters:
  - min_mass: Minimum mass value of the IMF range (float or astropy.units.Quantity).
  - max_mass: Maximum mass value of the IMF range (float or astropy.units.Quantity).
  - mass_function: Mass function to use for the IMF (optional).
  - unit: Unit of mass (optional, defaults to solar masses).
  - number_samples: Number of samples to use for interpolating the CDF (default: 100).
  - seed: Value to seed the random number generator with (optional, int, or None to turn off).
  """

  def __init__(self, min_mass, max_mass, mass_function=None, unit=units.solMass, number_samples=100, seed=None):
    self._number_samples = int(number_samples)
    self._seed = seed
    self.unit = unit if isUnit(unit) else units.solMass

    # Convert min_mass and max_mass to specified unit if they are Quantity objects, otherwise assume they are already in the correct unit
    self._min_mass = min_mass.to(self.unit) if isQuantity(min_mass) else min_mass * self.unit
    if self._min_mass.value <= 0: raise Exception('Cannot have minimum mass value be less than or equal to 0.')
    self._max_mass = max_mass.to(self.unit) if isQuantity(max_mass) else max_mass * self.unit

    # Determine the probability distribution function (PDF) based on the given mass function or default to a piecewise Chabrier 2003 and Salpeter 1955
    if mass_function is None: mass_function = lambda x: _numpy.where(x < 1, chabrier_2003_single(x), salpeter_1955(x, chabrier_2003_single(1)))
    self._imf = mass_function

    # Recalculate the IMF properties based on the updated parameters
    self._recalculate()

  def _recalculate(self):
    """
    Recalculates the IMF properties based on the current parameters.
    This function updates the normalization factor, normalized PDF, cumulative distribution function (CDF),
    mass values, and IMF values based on the current min_mass, max_mass, and PDF.
    """
    # Calculate the normalization factor for the PDF
    pdf = _numpy.vectorize(lambda x: _numpy.where(x < self.min_mass.value, 0, _numpy.where(x > self.max_mass.value, 0, self._imf(x))))
    normalization_factor = _quad(pdf, self._min_mass.value, self._max_mass.value)[0]

    # Create a normalized PDF
     
    npdf = lambda x: pdf(x) / normalization_factor
    self._npdf = _numpy.vectorize(npdf)

    # Create a cumulative distribution function (CDF)
    cdf = lambda x: _numpy.where(x < self.min_mass.value, 0, _numpy.where(x > self.max_mass.value, 1,  _quad(self._npdf, self._min_mass.value, x)[0]))
    self._cdf = _numpy.vectorize(cdf)

    # Generate logarithmically spaced mass values between min_mass and max_mass
    self._masses = _numpy.logspace(_numpy.log10(self._min_mass.value), _numpy.log10(self._max_mass.value), self.number_samples)

    # Calculate the CDF and IMF values for the mass values
    self._CDF = self._cdf(self._masses)
    self._norm_imf = self._npdf(self._masses)

  def random_mass(self, size=1, **kwargs):
    """
    Generates random mass values from the IMF.

    Parameters:
    - size: Number of random mass values to generate (default: 1).

    Returns:
    - Randomly generated mass value(s) from the IMF.
    """
    if 'seed' in kwargs: self.seed = kwargs['seed']
    if self.seed != None: _numpy.random.seed(self.seed)
    rand_masses = _numpy.interp(_numpy.random.uniform(size=size), self._CDF, self._masses) * self.unit
    if size > 1: return rand_masses
    else: return rand_masses[0]

  def masses(self, number_samples, endpoint=True):
    """
    Generates an array of mass values logarithmically spanning the IMF range.

    Parameters:
    - number_samples: Number of mass values to generate.

    Returns:
    - numpy array of mass values logarithmically spanning the IMF range.
    """
    return _numpy.logspace(_numpy.log10(self._min_mass.value), _numpy.log10(self._max_mass.value), number_samples, endpoint=endpoint)

  @property
  def min_mass(self):
    """
    Minimum mass value of the IMF range.
    """
    return self._min_mass

  @min_mass.setter
  def min_mass(self, value):
    """
    Setter for the minimum mass value of the IMF range.
    Recalculates the IMF properties when the min_mass value is updated.

    Parameters:
    - value: New minimum mass value (float or astropy.units.Quantity).
    """
    value = value.to(self.unit) if isQuantity(value) else value * self.unit
    if value.value <= 0: raise Exception('Cannot have minimum mass value be less than or equal to 0.')
    self._min_mass = value
    self._recalculate()

  @property
  def max_mass(self):
    """
    Maximum mass value of the IMF range.
    """
    return self._max_mass

  @max_mass.setter
  def max_mass(self, value):
    """
    Setter for the maximum mass value of the IMF range.
    Recalculates the IMF properties when the max_mass value is updated.

    Parameters:
    - value: New maximum mass value (float or astropy.units.Quantity).
    """
    self._max_mass = value.to(self.unit) if isQuantity(value) else value * self.unit
    self._recalculate()

  @property
  def median_mass(self):
    """
    Median mass value of the IMF.
    """
    return _numpy.interp(0.5, self._CDF, self._masses) * self.unit
  
  @property
  def seed(self):
    """
    Seed for the random number generator.
    """
    return self._seed

  @seed.setter
  def seed(self, value):
    """
    Setter for the seed for the random number generator 

    Parameters:
    - value: New seed for the random number generator (int, or None to turn off)
    """
    self._seed = value

  @property
  def number_samples(self):
    """
    Seed for the random number generator.
    """
    return self._number_samples

  @number_samples.setter
  def number_samples(self, value):
    """
    Setter for the seed for the random number generator 

    Parameters:
    - value: New seed for the random number generator (int, or None to turn off)
    """
    self._number_samples = int(value)
    self._recalculate()

  @property
  def cdf(self):
    """
    Cumulative distribution function (CDF) of the IMF.
    """
    return lambda x: _numpy.interp(x, self._masses, self._CDF)

  @property
  def CDF(self):
    """
    Cumulative distribution function (CDF) of the IMF.
    """
    return self.cdf

  @property
  def pdf(self):
    """
    Probability density function (PDF) of the IMF.
    """
    return self._npdf

  @property
  def PDF(self):
    """
    Probability density function (PDF) of the IMF.
    """
    return self.pdf

  @property
  def imf(self):
    """
    Probability density function (PDF) of the IMF.
    """
    return self._imf

  @property
  def IMF(self):
    """
    Probability density function (PDF) of the IMF.
    """
    return self.imf



from lib2to3.pytree import convert
import numpy as _numpy
import rebound as _rebound
import astropy.constants as const
from . import units

twopi = 2.*_numpy.pi

class UnitSystem():

  def __init__(self, UNIT_SYSTEM=[]) -> None:
    self._units = {'length': units.au, 'time': units.Myr, 'mass': units.solMass, 'angle': units.rad, 'velocity': units.km/units.s, 'object': units.stars, 'density': units.stars/units.pc**3}
    self.UNIT_SYSTEM = UNIT_SYSTEM
    pass

  @property
  def units(self):
    return self._units

  @property
  def UNIT_SYSTEM(self):
    return self._UNIT_SYSTEM

  @UNIT_SYSTEM.setter
  def UNIT_SYSTEM(self, UNIT_SYSTEM):
    if UNIT_SYSTEM != []:
      lengthUnit = [this for this in UNIT_SYSTEM if this.is_equivalent(units.m)]
      self._units['length'] = lengthUnit[0] if lengthUnit != [] else self._units['length']

      timeUnit = [this for this in UNIT_SYSTEM if this.is_equivalent(units.s)]
      self._units['time'] = timeUnit[0] if timeUnit != [] else self._units['time']

      velocityUnit = [this for this in UNIT_SYSTEM if this.is_equivalent(units.km/units.s)]
      if velocityUnit == [] and timeUnit != [] and lengthUnit != []: velocityUnit = [lengthUnit[0]/timeUnit[0]]
      self._units['velocity'] = velocityUnit[0] if velocityUnit != [] else self._units['velocity']

      massUnit = [this for this in UNIT_SYSTEM if this.is_equivalent(units.kg)]
      self._units['mass'] = massUnit[0] if massUnit != [] else self._units['mass']

      angleUnit = [this for this in UNIT_SYSTEM if this.is_equivalent(units.rad)]
      self._units['angle'] = angleUnit[0] if angleUnit != [] else self._units['angle']

      objectUnit = [this for this in UNIT_SYSTEM if this.is_equivalent(units.stars)]
      self._units['object'] = objectUnit[0] if objectUnit != [] else units.stars

      densityUnit = [this for this in UNIT_SYSTEM if this.is_equivalent(units.stars/units.m**3)]
      densityUnit2 = [this for this in UNIT_SYSTEM if this.is_equivalent(1/units.m**3)]
      if densityUnit == [] and densityUnit2 != []: densityUnit = [self._object_unit * densityUnit2[0]]
      elif densityUnit == [] and objectUnit != [] and lengthUnit != []: densityUnit = [self._units['object']/self._units['length']**3]
      self._units['density'] = densityUnit[0] if densityUnit != [] else self._units['density']
    
    self._UNIT_SYSTEM = list(self._units.values())



############################################################
################### Helper Functions #######################
############################################################

def rebound_units(sim):
    defrebunits = {'length': units.au, 'mass': units.solMass, 'time': units.yr2pi}
    siunits = {'length': units.m, 'mass': units.kg, 'time': units.s}
    rebunitset = {'length': _rebound.units.lengths_SI, 'mass': _rebound.units.masses_SI, 'time': _rebound.units.times_SI}
    simunits = sim.units
    
    for unit in simunits:
        if simunits[unit] == None: simunits[unit] = defrebunits[unit]
        else: simunits[unit] = rebunitset[unit][simunits[unit]] * siunits[unit]
    return simunits

# Implemented from StackOverflow: https://stackoverflow.com/a/14314054
def moving_average(a, n=3) :
    '''Compute the moving average of an array of numbers using the nearest n elements.'''
    ret = _numpy.cumsum(a)
    ret[n:] = ret[n:] - ret[:-n]
    return ret[n - 1:] / n


def vinf_and_b_to_e(mu, star_b, star_vinf):
    '''
        Using the impact parameter to convert from the relative velocity at infinity between the two stars to the eccentricity of the flyby star.
        Equation (2) from Spurzem et al. (2009) https://ui.adsabs.harvard.edu/abs/2009ApJ...697..458S/abstract

        Parameters
        ----------
        mu : the total mass of the system (Sun, planets, and flyby star) times the gravitational constant G
        b :  impact parameter of the flyby star
        vinf : the relative velocity at infinity between the central star and the flyby star (hyperbolic excess velocity)
    '''

    star_b = star_b.to(units.au) if isQuantity(star_b) else star_b * units.au
    star_vinf = star_vinf.to(units.km/units.s) if isQuantity(star_vinf) else star_vinf * units.km/units.s

    numerator = star_b * star_vinf**2.
    return _numpy.sqrt(1 + (numerator/mu)**2.)

def determine_eccentricity(sim, star_mass, star_b, star_v=None, star_e=None):
    '''Calculate the eccentricity of the flyby star. '''
    # Convert the units of the REBOUND Simulation into Astropy Units.
    units = rebound_units(sim)
    G = (sim.G * units['length']**3 / units['mass'] / units['time']**2)
    star_mass = star_mass.to(units['mass']) if isQuantity(star_mass) else star_mass * units['mass']
    star_b = star_b.to(units['length']) if isQuantity(star_b) else star_b * units['length']
    star_v = star_v.to(units['length']/units['time']) if isQuantity(star_v) else star_v * units['length']/units['time']

    mu = G * (_numpy.sum([p.m for p in sim.particles]) * units['mass'] + star_mass)
    if star_e is not None and star_v is not None: raise AssertionError('Overdetermined. Cannot specify an eccentricity and a velocity for the perturbing star.')
    elif star_e is not None and star_v is None:
        # Simply use the eccentricity if it is defined.
        return star_e
    elif star_e is None and star_v is not None:
        # If `star_v` is defined convert it to eccentricity.
        # Assumes that `star_v` is in units of km/s.
        return vinf_and_b_to_e(mu=mu, star_b=star_b, star_vinf=star_v)
    else: raise AssertionError('Undetermined. Specify either an eccentricity or a velocity for the perturbing star.')

def maxwell_boltzmann_scale_from_dispersion(sigma):
    '''
        Converts velocity dispersion (sigma) to scale factor for Maxwell-Boltzmann distributions.
    '''
    return _numpy.sqrt((_numpy.pi*_numpy.square(sigma))/(3.0*_numpy.pi - 8.0))

def maxwell_boltzmann_scale_from_mean(mu):
    '''
        Converts mean (mu) to scale factor for Maxwell-Boltzmann distributions.
    '''
    return _numpy.sqrt(_numpy.pi/2.0) * (mu / 2.0)

def maxwell_boltzmann_mean_from_dispersion(sigma):
    '''
        Converts velocity dispersion (sigma) to mean (mu) for Maxwell-Boltzmann distributions.
    '''
    scale = maxwell_boltzmann_scale_from_dispersion(sigma)
    return (2.0 * scale) * _numpy.sqrt(2.0/_numpy.pi)

def maxwell_boltzmann_mode_from_dispersion(sigma):
    '''
        Converts velocity dispersion (sigma) to mode (most common or typical value) for Maxwell-Boltzmann distributions.
    '''
    scale = maxwell_boltzmann_scale_from_dispersion(sigma)
    return scale * _numpy.sqrt(2.0)

def cross_section(M, R, v, unit_set=UnitSystem()):
    '''
        The cross-section with gravitational focusing.
        
        Parameters
        ----------
        M : the mass of flyby star (default units: solMass)
        R : the maximum interaction radius (default units: AU)
        v : the typical velocity from the distribution (default units: km/s)
    '''

    # Newton's gravitational constant in units of Msun, AU, and Years/2pi (G ~ 1).
    G = const.G.decompose(unit_set.UNIT_SYSTEM) 
    sun_mass = 1 * units.solMass # mass of the Sun in units of Msun

    v = verify_unit(v, unit_set.units['velocity'])
    R = verify_unit(R, unit_set.units['length'])
    M = verify_unit(M, unit_set.units['mass'])

    return (_numpy.pi * R**2) * (1 + 2*G*(sun_mass + M)/(R * v**2))

def encounter_rate(n, v, R, M=(1 * units.solMass), unit_set=UnitSystem()):
    '''
        The expected flyby encounter rate within an stellar environment
        
        Parameters
        ----------
        n : stellar number density (default units: pc^{-3})
        v : average velocity  (default units: km/s)
        R : interaction radius (default units: AU)
        M : mass of a typical flyby star (default units: solMass)
    '''
    n = verify_unit(n, unit_set.units['density'])
    v = verify_unit(v, unit_set.units['velocity'])
    R = verify_unit(R, unit_set.units['length'])
    M = verify_unit(M, unit_set.units['mass'])
    
    return n * v * cross_section(M, R, v, unit_set)

def verify_unit(value, unit):
    return value.to(unit) if isQuantity(value) else value * unit

def isList(l):
    '''Determines if an object is a list or numpy array. Used for flyby parallelization.'''
    if isinstance(l,(list,_numpy.ndarray)): return True
    else: return False


def isQuantity(var):
    '''Determines if an object is an Astropy Quantity. Used for Stellar Environment initializations.'''
    return isinstance(var, units.quantity.Quantity)

def isUnit(var):
    '''Determines if an object is an Astropy Quantity. Used for Stellar Environment initializations.'''
    return isinstance(var, (units.core.IrreducibleUnit, units.core.CompositeUnit))
